"""Module loading agents' movement"""
import csv

from PySide2.QtCore import Property, QObject, QUrl, Signal, Slot
from PySide2.QtGui import QColor


class Color:
    """Class calculating agent color."""

    def __init__(self, h, s, l, a):
        """Initialize with base HSLA color."""
        self._color = (h, s, l, a)

    def str(self):
        """Get color in hsla() formatted string."""
        h, s, l, a = self._color
        return f"hsla({h}, {s}%, {l}%, {a})"

    def obj(self):
        """Get Qt color object."""
        h, s, l, a = self._color
        return QColor.fromHsl(360 * h / 360, 255 * s / 100, 255 * l / 100, 255 * a)


class AgentManager(QObject):
    """Class controlling display and keyframing of picking agents."""

    agentsChanged = Signal()
    activeAgentsChanged = Signal()

    def __init__(self):
        """Initialize agent manager."""
        QObject.__init__(self)
        self.agents = {}
        self.keys = set()
        self._max_time = 0
        self._active = False

    def _sorted_keys(self):
        """Sort keys for displaying ALL agents in list in nice order."""
        try:
            return map(str, sorted(map(int, self.agents.keys())))
        except Exception:
            return sorted(self.agents.keys())

    @Property(int, constant=False, notify=agentsChanged)
    def max_time(self):
        """Get maximal time."""
        return self._max_time

    @Property(bool, constant=False, notify=agentsChanged)
    def active(self):
        """Get true if agent visualization is active false otherwise."""
        return self._active

    @Property("QVariantList", constant=False, notify=agentsChanged)
    def agent_list(self):
        """Get description of agents for list visualization."""
        agents = []
        for k in self._sorted_keys():
            a = self.agents[k]
            agents.append(
                {
                    "name": k,
                    "color": a["color"].obj(),
                    "steps": len(a["steps"]),
                    "checked": k in self.keys,
                }
            )
        return agents

    @Slot(str, bool)
    def toggle_agent(self, key, checked):
        """Toggle agent from active agents set.

        Args:
            key (str): name/key of agent to toggle.
        """
        if checked:
            self.keys.add(key)
        elif key in self.keys:
            self.keys.remove(key)
        self.activeAgentsChanged.emit()

    @Slot(result="QVariantList")
    def get_colors(self):
        """Get list of colors for displaying agents."""
        return [self.agents[k]["color"].str() for k in sorted(self.keys)]

    @Slot(int, result="QVariantList")
    def get_timestep(self, time):
        """Get representation of positions for given time.

        Args:
            time (int): time of step (0 <= time <= max_time)

        Returns:
            list[list[int]]: list of agent steps
        """
        steps = []
        for k in sorted(self.keys):
            agent = self.agents[k]
            if agent["offset"] <= time < agent["offset"] + len(agent["steps"]):
                steps.append(agent["steps"][time - agent["offset"]])
            else:
                steps.append([-1, -1, 0, 0])
        return steps

    @Slot(QUrl)
    def load_data(self, file_path):
        """Load data picker agents data from provided CSV file.

        Args:
            file_path (QUrl): file path to CSV file
        """
        # First clean the list
        self.agents = {}
        self.agentsChanged.emit()

        with open(file_path.toLocalFile(), newline="") as csvfile:
            reader = csv.DictReader(csvfile)
            for r in reader:
                agent = r["agent"]
                if agent not in self.agents:
                    self.agents[agent] = {
                        "offset": int(r["time"]),
                        "steps": [
                            list(
                                map(
                                    int,
                                    (r["locX"], r["locY"], r["isPick"], r["orderId"]),
                                )
                            )
                        ],
                    }
                else:
                    time_diff = int(r["time"]) - (
                        self.agents[agent]["offset"] + len(self.agents[agent]["steps"])
                    )
                    if time_diff > 0:
                        self.agents[agent]["steps"].extend(
                            time_diff * [self.agents[agent]["steps"][-1]]
                        )
                    self.agents[agent]["steps"].append(
                        list(
                            map(int, (r["locX"], r["locY"], r["isPick"], r["orderId"]))
                        )
                    )

        for i, k in enumerate(self.agents.keys()):
            self.agents[k]["color"] = Color(
                int(360 * i / len(self.agents)), 75, 70, 1.0
            )

        self._max_time = max(
            self.agents[a]["offset"] + len(self.agents[a]["steps"]) - 1
            for a in self.agents
        )
        self.keys = set(self.agents.keys())
        self._active = True
        self.agentsChanged.emit()



"""Utils for parsing the Excel data."""
from datetime import datetime

dim_factors = {
    "m": 1,
    "meters": 1,
    "dm": 0.1,
    "cm": 0.01,
    "mm": 0.001,
}

weight_factors = {
    "kg": 1,
    "g": 0.001,
}

# Mapping possible names to standard types
types = {
    "floor": "floor",
    "rack": "rack",
    "storage rack": "rack",
    "wall": "wall",
    "inbound door": "inbound_door",
    "outbound door": "outbound_door",
    "staging area": "staging_area",
    "custom": "custom",
}


def estimate_sheet_type(sheet):
    """Estimate sheet type based on sheet name."""
    sheet = sheet.lower()
    if "coord" in sheet:
        return "Coordinates"
    if "loc" in sheet:
        return "Locations"
    if "item" in sheet:
        return "Items"
    if "inv" in sheet:
        return "Inventory"
    if "ord" in sheet:
        return "Orders"
    return "None"


def convert_type(type_str):
    """Unify location type names."""
    return types.get(" ".join(type_str.lower().split()), "custom")


def convert_dim(dim, uom):
    """Convert dimension size to meters."""
    if uom:
        return dim * dim_factors[uom.lower()]
    return dim


def convert_weight(weight, uom):
    """Convert weight to kilograms."""
    if uom:
        return weight * weight_factors[uom.lower()]
    return weight


def convert_date(date, fmt="%d.%m.%Y"):
    """Convert string date to datetime."""
    if type(date) is datetime:
        return date
    if date:
        return datetime.strptime(date, fmt)
    return None



"""Module with plug-in for calculating selected items frequencies."""

from virtual_warehouse.plugin import BasePlugin


class Plugin(BasePlugin):
    """Sub-class of base plugin providing calculation of frequency.
    Calculate frequency as frequency of item in order.
    """

    display_name = "&Order Histogram"

    def __init__(self, locations, items, orders, inventory):
        """Initialize plugin and save required parameters.

        Args:
            locations (dict): dictionary of locations mapping id to locations
            items (dict): dictionary of items mapping id to items
            orders (dict): dictionary of orders mapping id to order
            inventory (dict): dictionary of specifying inventory
        """
        super().__init__(locations, items, orders, inventory)
        self.orders = orders
        self.item_locs = Plugin.item_locations(inventory)
        self.date = list(self.item_locs.keys())[-1]

    @staticmethod
    def item_locations(inventory):
        """Create dictionary mapping (date, item) to list of locations."""
        item_to_loc = {}
        for date, status in inventory.items():
            item_to_loc[date] = {}
            for loc_inventories in status.values():
                for inv in loc_inventories:
                    if inv.has_item.name not in item_to_loc[date]:
                        item_to_loc[date][inv.has_item.name] = []
                    item_to_loc[date][inv.has_item.name].append(inv.has_location)

        return item_to_loc

    def on_orders_update(self, clear, add, ids):
        """Update frequency calculation on orders check/uncheck.

        Args:
            clear (bool): Whether to clear previously selected orders.
            add (bool): Whether to add or remove ids from calculation.
            ids (list[str]): list of ids considered in calculation.
        """
        if clear:
            self._clear_frequencies()

        for order_id in ids:
            for ord_item in self.orders[order_id].has_ordered_items:
                for loc in self.item_locs[self.date][ord_item.has_item.name]:
                    loc.has_freq += (1 if add else -1) * ord_item.has_total_qty

    def _calculate_freq(self, locations, items, orders):
        """Calculate frequency for individual locations based on selected orders."""
        self.on_orders_update(False, True, orders)



import importlib
import pkgutil
from abc import ABC, abstractmethod

from PySide2.QtCore import Property, QObject, Qt, Signal


class BasePlugin(ABC):
    """Base plugin class for calculating frequencies.
    Each plugin must be subclass of this BasePlugin class.
    Plugins must be placed as separate modules in virtual_warehouse.plugins
    Implementation of (_calculate_freq(...)) initial frequency calculation is required.
    Plugin can also implement some of following functions to dynamically update
    frequencies:
        on_locations_update(...)
        on_items_update(...)
        on_orders_update(...)


    Attributes:
        locations (dict[str, Location]): always store dictionary of locations
        display_name (str): name of plugin to display in Menu (must be overwritten)
    """

    display_name: str

    def __init__(self, locations, items, orders, inventory):
        """Initialize plugin, all plugins must implement constructor with same arguments.

        Args:
            locations (dict): dictionary of locations mapping id to locations
            items (dict): dictionary of items mapping id to items
            orders (dict): dictionary of orders mapping id to order
            inventory (dict): dictionary of specifying inventory
        """
        self.locations = locations

    def calculate_frequencies(self, *args):
        """Run initial frequency calculation."""
        self._clear_frequencies()
        self._calculate_freq(*args)

    def on_locations_update(self, clear, add, ids):
        """Update frequency calculation on locations check/uncheck.
        By default this function does nothing. Should be implemented by plugin,
        if plugin frequencies change with location selection.

        Args:
            clear (bool): Whether to clear previously selected orders.
            add (bool): Whether to add or remove ids from calculation.
            ids (list[str]): list of ids considered in calculation.
        """

    def on_items_update(self, clear, add, ids):
        """Update frequency calculation on items check/uncheck.
        By default this function does nothing. Should be implemented by plugin,
        if plugin frequencies change with item selection.

        Args:
            clear (bool): Whether to clear previously selected orders.
            add (bool): Whether to add or remove ids from calculation.
            ids (list[str]): list of ids considered in calculation.
        """

    def on_orders_update(self, clear, add, ids):
        """Update frequency calculation on orders check/uncheck.
        By default this function does nothing. Should be implemented by plugin,
        if plugin frequencies change with order selection.

        Args:
            clear (bool): Whether to clear previously selected orders.
            add (bool): Whether to add or remove ids from calculation.
            ids (list[str]): list of ids considered in calculation.
        """

    @abstractmethod
    def _calculate_freq(self, locations, items, orders):
        """Calculate frequencies based on selected objects (abstract method).

        Args:
            locations (list[str]): list of selected location ids
            items (list[str]): list of selected item ids
            orders (list[str]): list of selected order ids
        """

    def _clear_frequencies(self):
        """Clear all frequencies."""
        for v in self.locations.values():
            v.has_freq = 0


class PluginManager(QObject):
    """Class loading and controlling all plugins and providing them to controller."""

    pluginChanged = Signal()

    def __init__(self, location_model, item_model, order_model, update_signal):
        """Initialize PluginManager holding all plugins from plugins folder.

        Args:
            location_model (UniversalListModel): list model representing location tab
            item_model (UniversalListModel): list model representing item tab
            order_model (UniversalListModel): list model representing order tab
            update_signal (Signal): signal to emit on plugin update (drawModeChanged)
        """
        QObject.__init__(self)
        self.plugin_modules = PluginManager.load_plugins()
        self.plugins = {}
        self.active_plugin = None
        self._is_active = False

        self._location_model = location_model
        self._item_model = item_model
        self._order_model = order_model

        self.update_signal = update_signal

        self._location_model.checkChanged.connect(
            self._locations_update, Qt.DirectConnection
        )
        self._item_model.checkChanged.connect(self._items_update, Qt.DirectConnection)
        self._order_model.checkChanged.connect(self._orders_update, Qt.DirectConnection)

    def set_data(self, locations, items, orders, inventory):
        """Set new warehouse data for all plugins."""
        if (
            locations is not None
            and items is not None
            and orders is not None
            and inventory is not None
        ):
            for name, module in self.plugin_modules.items():
                self.plugins[name] = module.Plugin(locations, items, orders, inventory)
            self._is_active = True
            self._update()

    @Property(str, constant=False, notify=pluginChanged)
    def active(self):
        """Get name of active plugin."""
        return self.active_plugin

    @active.setter
    def activate_plugin(self, plugin_name):
        """Activate selected plugin and update frequencies.

        Args:
            plugin_name (str): plugin name works as key for plugins dictionary
        """
        if self.active_plugin != plugin_name:
            self.active_plugin = plugin_name
            self._update()
            self.pluginChanged.emit()

    def _locations_update(self, args):
        """Update active plugin on locations update."""
        if self.active_plugin and (args[0] or len(args[2]) != 0):
            self.plugins[self.active_plugin].on_locations_update(*args)
            self.update_signal.emit()

    def _items_update(self, args):
        """Update active plugin on items update."""
        if self.active_plugin and (args[0] or len(args[2]) != 0):
            self.plugins[self.active_plugin].on_items_update(*args)
            self.update_signal.emit()

    def _orders_update(self, args):
        """Update active plugin on orders update."""
        if self.active_plugin and (args[0] or len(args[2]) != 0):
            self.plugins[self.active_plugin].on_orders_update(*args)
            self.update_signal.emit()

    def _update(self):
        """Recalculate frequencies using active plugin."""
        if self.active_plugin and self._is_active:
            self.plugins[self.active_plugin].calculate_frequencies(
                self._location_model.checked,
                self._item_model.checked,
                self._order_model.checked,
            )

        if self._is_active:
            self.update_signal.emit()

    @Property("QVariantList", constant=True)
    def names(self):
        """Get list of display names and module names."""
        return [
            {"name": m.Plugin.display_name, "module": k}
            for k, m in self.plugin_modules.items()
        ]

    @staticmethod
    def _iter_namespace(ns_pkg):
        """Iterate modules in specified sub-package."""
        return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")

    @staticmethod
    def load_plugins():
        """Load plugin modules from virtual_warehouse.plugins sub-package.

        Returns:
            dict[str, module]: dictionary containing module name and module.
        """
        import virtual_warehouse.plugins

        return {
            name.split(".")[-1]: importlib.import_module(name)
            for finder, name, ispkg in PluginManager._iter_namespace(
                virtual_warehouse.plugins
            )
        }



"""Parser of Excel data files."""
from datetime import datetime

import xlsxio
from xlrd import open_workbook

from virtual_warehouse.data.data_model import (
    Inventory,
    Item,
    ItemUnit,
    Location,
    Order,
    RackLocation,
)
from virtual_warehouse.data.utils import convert_date, convert_type, estimate_sheet_type


class Document:
    """Document class which loads xls or xlsx file and parse different data objects."""

    def __init__(self, file_path):
        # Determines backend for loading documents (xlsx files uses openpyxl)
        self.is_xlsx = Document.check_xlsx(file_path)
        self.locations = {}
        self.items = {}
        self.balance = {}
        self.orders = {}
        if self.is_xlsx:
            self.doc = xlsxio.XlsxioReader(file_path)
        else:
            self.doc = open_workbook(file_path)

    @staticmethod
    def check_xlsx(file_path):
        """Check if document is .xlsx document (required for openpyxl library)."""
        return file_path[-5:] == ".xlsx"

    @staticmethod
    def get_sheet_names(file_path):
        """Get names of all sheets in document."""
        if Document.check_xlsx(file_path):
            with xlsxio.XlsxioReader(file_path) as reader:
                names = reader.get_sheet_names()
        else:
            doc = open_workbook(file_path, on_demand=True)
            names = doc.sheet_names()
        return [[n, estimate_sheet_type(n)] for n in names]

    def close(self):
        """Release resources owned by the document."""
        if self.is_xlsx:
            self.doc.close()

    def parse_locations(self, sheet_name="LOCATIONmaster"):
        """Parse LOCATIONmaster sheet."""
        if self.is_xlsx:
            types = [str, str, str, str, float, float, float, str, float, str, str]
            with self.doc.get_sheet(sheet_name, types=types) as sheet:
                sheet.read_header()
                for row in sheet.iter_rows():
                    if len(row) == 0 or not row[0]:
                        continue

                    location_id = row[0]
                    if convert_type(row[1]) == "rack":
                        self.locations[location_id] = RackLocation.create(*row[:11])
                    else:
                        self.locations[location_id] = Location.create(*row[:11])

        else:
            sheet = self.doc.sheet_by_name(sheet_name)

            for row in range(1, sheet.nrows):
                location_id = str(sheet.cell(row, 0).value)
                if not location_id:
                    continue

                values = [sheet.cell(row, i).value for i in range(11)]
                if convert_type(values[1]) == "rack":
                    self.locations[location_id] = Location.create(*values)
                else:
                    self.locations[location_id] = RackLocation.create(*values)

        return self.locations

    def parse_coordinates(self, sheet_name="XYZ_coordinates"):
        """Parse XYZ_coordinates sheet."""
        if self.is_xlsx:
            types = [str, float, float, float]
            with self.doc.get_sheet(sheet_name, types=types) as sheet:
                sheet.read_header()
                for row in sheet.iter_rows():
                    if len(row) == 0 or not row[0]:
                        continue
                    location_id = row[0]
                    self.locations[location_id].set_coord(*row[1:4])

        else:
            sheet = self.doc.sheet_by_name(sheet_name)

            for row in range(1, sheet.nrows):
                location_id = str(sheet.cell(row, 0).value)
                if not location_id:
                    continue

                self.locations[location_id].set_coord(
                    *(sheet.cell(row, i).value for i in range(1, 4))
                )

        return self.locations

    def parse_items(self, sheet_name="ITEMmaster"):
        """Parse ITEMmaster sheet."""
        if self.is_xlsx:
            types = [str, str, str, str] + 5 * [
                int,
                str,
                float,
                float,
                float,
                str,
                float,
                str,
            ]
            with self.doc.get_sheet(sheet_name, types=types) as sheet:
                sheet.read_header()
                for row in sheet.iter_rows():
                    if len(row) == 0 or not row[0]:
                        continue

                    item_id, description, gtype, zone = row[:4]
                    unit_levels = []
                    for col in range(4, len(row), 8):
                        unit_levels.append(
                            ItemUnit.create(f"{item_id}-u{col}", *row[col : col + 8],)
                        )
                    self.items[item_id] = Item.create(
                        item_id, description, gtype, zone, unit_levels[0], unit_levels
                    )

        else:
            sheet = self.doc.sheet_by_name(sheet_name)

            for row in range(1, sheet.nrows):
                item_id, description, gtype, zone = (
                    sheet.cell(row, i).value for i in range(4)
                )
                item_id = str(item_id)
                if not item_id:
                    continue

                unit_levels = []
                for col in range(4, sheet.ncols, 8):
                    unit_levels.append(
                        ItemUnit.create(
                            f"{item_id}-u{col}",
                            *(sheet.cell(row, col + i).value for i in range(8)),
                        )
                    )
                self.items[item_id] = Item.create(
                    item_id, description, gtype, zone, unit_levels[0], unit_levels
                )

        return self.items

    def parse_inventory_balance(self, sheet_name="Inventory Ballance"):
        """Parse Inventory Balance sheet  ('balance' in final version, most likely)."""
        if self.is_xlsx:
            types = [datetime, str, str, str, datetime, int, int, int, int, int]
            with self.doc.get_sheet(sheet_name, types=types) as sheet:
                sheet.read_header()
                for row in sheet.iter_rows():
                    if len(row) == 0 or not row[0]:
                        continue
                    date, location_id = row[:2]
                    if date not in self.balance:
                        self.balance[date] = {location_id: []}
                    elif location_id not in self.balance[date]:
                        self.balance[date][location_id] = []

                    self.balance[date][location_id].append(Inventory.create(*row[:10]))

        else:
            sheet = self.doc.sheet_by_name(sheet_name)

            for row in range(1, sheet.nrows):
                date = convert_date(sheet.cell(row, 0).value, "%d.%m.%Y")
                location_id = str(sheet.cell(row, 1).value)
                if not date:
                    continue

                if date not in self.balance:
                    self.balance[date] = {location_id: []}
                elif location_id not in self.balance[date]:
                    self.balance[date][location_id] = []

                self.balance[date][location_id].append(
                    Inventory.create(*(sheet.cell(row, i).value for i in range(10)))
                )

        return self.balance

    def parse_orders(self, sheet_name="Order"):
        """Parse Order sheet."""
        if self.is_xlsx:
            types = [str, str, str, str, str, str, int, str, int, int, str]
            with self.doc.get_sheet(sheet_name, types=types) as sheet:
                sheet.read_header()
                for row in sheet.iter_rows():
                    if len(row) == 0 or not row[0]:
                        continue

                    order_id = row[0]
                    if order_id in self.orders:
                        self.orders[order_id].add_item(*row[7:11])
                    else:
                        self.orders[order_id] = Order.create(*row[:11])

        else:
            sheet = self.doc.sheet_by_name(sheet_name)

            for row in range(1, sheet.nrows):
                order_id = str(sheet.cell(row, 0).value)
                if not order_id:
                    continue

                if order_id in self.orders:
                    self.orders[order_id].add_item(
                        *(sheet.cell(row, i).value for i in range(7, 11))
                    )
                else:
                    self.orders[order_id] = Order.create(
                        *(sheet.cell(row, i).value for i in range(11))
                    )

        return self.orders

    def parse_document(self):
        """Parse the whole document."""
        locations = self.parse_locations()
        locations = self.parse_coordinates()
        items = self.parse_items()
        balance = self.parse_inventory_balance()
        orders = self.parse_orders()

        return locations, items, balance, orders



"""Module providing content for sidebar tabs: locations/items/orders.
It provides classes wrapping the location/item/order objects along with list model.
It also provides functionality for checking different items.

HoverListModel is for displaying side view of location which is selected or hovered.
"""
import locale

from PySide2.QtCore import (
    Property,
    QAbstractListModel,
    QModelIndex,
    QObject,
    Qt,
    Signal,
    Slot,
)

locale.setlocale(locale.LC_ALL, "")


class TabLocation(QObject):
    """Location QObject for displaying list of selected locations."""

    def __init__(self, location):
        QObject.__init__(self)
        self._i = location
        self._checked = False

    checkedChanged = Signal()

    @Property(str, constant=True)
    def name(self):
        """Get name of location (property)."""
        return self._i.name

    @Property(str, constant=True)
    def class_str(self):
        """Get class + subclass representation (property)."""
        if self._i.has_lsubclass:
            return f"{self._i.has_lclass} > {self._i.has_lsubclass}"
        return self._i.has_lclass

    @Property(str, constant=True)
    def dimension_str(self):
        """Get dimension representation (property)."""
        return f"{self._i.has_length} x {self._i.has_width} x {self._i.has_height}"

    @Property(int, constant=True)
    def max_weight(self):
        """Get maximal weight supported by location (property)."""
        return self._i.has_max_weight

    @Property(str, constant=True)
    def zone(self):
        """Get zone where is location located (property)."""
        return self._i.has_zone

    @Property(str, constant=True)
    def z(self):
        """Get z - vertical direction - coordinate of location (property)."""
        return str(int(self._i.has_z))

    @Property(float, constant=True)
    def heat(self):
        """Calculate heat for side-view."""
        return self._i.has_freq

    @Property(bool, constant=False, notify=checkedChanged)
    def checked(self):
        """Get checked state of location (property)."""
        return self._checked

    @checked.setter
    def set_checked(self, val):
        """Set checked state of location (property setter)."""
        self._checked = val
        self.checkedChanged.emit()


class TabItem(QObject):
    """Item QObject for displaying list of items in warehouse."""

    def __init__(self, item):
        QObject.__init__(self)
        self._i = item
        self._checked = False

    checkedChanged = Signal()

    @Property(str, constant=True)
    def name(self):
        """Get name of item (property)."""
        return self._i.name

    @Property(str, constant=True)
    def description(self):
        """Get description representation of item (property)."""
        if self._i.has_gtype:
            return f"{self._i.has_description} | {self._i.has_gtype}"
        return self._i.has_description

    @Property(str, constant=True)
    def zone(self):
        """Get zone which item requires (property)."""
        return self._i.has_zone

    @Property(str, constant=True)
    def base_dimension(self):
        """Get representation of base dimension of item package (property)."""
        bu = self._i.has_base_unit
        return f"{bu.has_length} x {bu.has_width} x {bu.has_height}"

    @Property(float, constant=True)
    def base_weight(self):
        """Get base weight of item package (property)."""
        return self._i.has_base_unit.has_weight

    @Property(bool, constant=False, notify=checkedChanged)
    def checked(self):
        """Get item checked state (property)."""
        return self._checked

    @checked.setter
    def set_checked(self, val):
        """Set item checked state (property setter)."""
        self._checked = val
        self.checkedChanged.emit()


class TabOrder(QObject):
    """Order QObject for displaying list of orders in warehouse."""

    def __init__(self, order):
        QObject.__init__(self)
        self._i = order
        self._checked = False

    checkedChanged = Signal()

    @Property(str, constant=True)
    def name(self):
        """Get name order (property)."""
        return str(self._i.name)

    @Property(str, constant=True)
    def direction(self):
        """Get direction of order (property)."""
        return self._i.has_direction

    @Property(str, constant=True)
    def date(self):
        """Get formatted date of order (property)."""
        return self._i.has_delivery_date.strftime("%x")

    @Property(str, constant=True)
    def item_id(self):
        """Get representation of item/items (property)."""
        if len(self._i.has_ordered_items) == 1:
            return str(self._i.has_ordered_items[0].has_item.name)
        return "Multi-item"

    @Property(int, constant=True)
    def num_items(self):
        """Get number of items in order."""
        return len(self._i.has_ordered_items)

    @Property(bool, constant=False, notify=checkedChanged)
    def checked(self):
        """Get order checked state (property)."""
        return self._checked

    @checked.setter
    def set_checked(self, val):
        """Set order checked state (property setter)."""
        self._checked = val
        self.checkedChanged.emit()


class UniversalListModel(QAbstractListModel):
    """Universal class for holding lists in sidebar tabs (locations, items, orders).

    Attributes:
        checkChanged (Signal): emits on new item in list checked.
            Emits update data (list[bool, bool, list[str]]) which represents:
            [is_clear, is_add, list_of_object_ids]. This is used to update plugins.
        filterChanged (Signal): emits on filter value changed

    """

    ObjectRole = Qt.UserRole + 1

    def __init__(self, object_class, objects=None, selected_objects=None, parent=None):
        """Initialize list model.

        Args:
            object_class (cls): class which wraps objects stored in list (TabItem,...)
            objects (dict[str, object]): dictionary of objects stored in list
            selected_objects (list[str]): list of ids (keys) of selected objects,
                selected objects are displayed in the list
            parent (object): required param for initialization
        """
        super(UniversalListModel, self).__init__(parent)
        self._object_class = object_class
        self._selected = [] if selected_objects is None else selected_objects
        self._all_selected = self._selected
        self.checked = set()
        self._search_text = ""
        self._filter = 0

        if objects is None:
            self._objects = {}
        else:
            self._objects = {k: self._object_class(v) for k, v in objects.items()}

    # Emits list as [is_clear, is_add, list_of_ids]
    checkChanged = Signal("QVariantList")
    filterChanged = Signal()

    def set_data(self, objects):
        """Set new objects and wrap then in wrapper class."""
        self._objects = {k: self._object_class(v) for k, v in objects.items()}

    def set_selected(self, selected, check=False):
        """Set new list of selected objects."""
        self._selected = selected
        self._all_selected = selected
        if check:
            self.set_checked(selected)
        self.layoutChanged.emit()

    def rowCount(self, parent=QModelIndex()):
        """Get number of items in list (required method)."""
        return len(self._selected)

    def data(self, index, role):
        """Return data for given role and index (required method)."""
        if index.isValid() and role == UniversalListModel.ObjectRole:
            return self._objects[self._selected[index.row()]]
        return None

    def roleNames(self):  # skipcq: PYL-R0201
        """Set role names, QAbstractListModel requires implementation of this method."""
        roles = {}
        roles[UniversalListModel.ObjectRole] = b"object"
        return roles

    def clear_checked(self):
        """Clear all checked items in list."""
        for n in self.checked:
            self._objects[n].set_checked(False)
        self.checked.clear()

    def set_checked(self, checked, control=False):
        """Set/add list of newly checked items in list.

        Args:
            checked (list[str]): list of ids of checked objects
            control (bool): if control is set to true, items are added to checked items
                otherwise originally checked items are cleared firs.
        """
        if not control:
            self.clear_checked()

        self.checked.update(checked)
        self._update_filter()
        for n in self.checked:
            self._objects[n].set_checked(True)

        self.checkChanged.emit([not control, True, checked])

    @Property(int, constant=False, notify=checkChanged)
    def check_state(self):
        """Return state of main checkbox of all visible check boxes.
        States: unchecked (0)/ partially checked (1) / checked (2)
        """
        selected_set = set(self._selected)
        match_count = sum(map(lambda x: x in selected_set, self.checked))
        if match_count:
            return 2 if match_count == len(self._selected) else 1
        return 0

    @Property(int, constant=False, notify=filterChanged)
    def filter(self):
        """Filter property for splitting tabs on all(0)/checked(1)/unchecked(2)."""
        return self._filter

    @filter.setter
    def set_filter(self, val):
        """Set value of filter property.

        Args:
            val (int): new value of filter (one of 0, 1, 2)
        """
        if self._filter != val:
            self._filter = val
            self.filterChanged.emit()
            self._update_filter()
            self.checkChanged.emit(
                [False, False, []]
            )  # Required for check_state update

    @Slot(bool)
    def check_all(self, check=True):
        """Check/Uncheck all currently visible list items."""
        self.set_checked(self._selected if check else [])

    @Slot(str, bool)
    def check(self, _id, check=True):
        """Check/uncheck one item in the list."""
        self._objects[_id].set_checked(check)
        if check:
            self.checked.add(_id)
        else:
            self.checked.remove(_id)

        self.checkChanged.emit([False, check, [_id]])

    @Slot(str)
    def search(self, text):
        """Search for text between object ids."""
        self._search_text = text
        self._update_filter()

    def _update_filter(self):
        """Filter items appearing in one of the lists: all/checked/unchecked/search."""
        if self._search_text:
            self._selected = [
                k for k in self._all_selected if self._search_text.lower() in k.lower()
            ]
        elif self._filter == 0:
            self._selected = self._all_selected
        elif self._filter == 1:
            self._selected = sorted(list(self.checked))
        elif self._filter == 2:
            self._selected = [k for k in self._all_selected if k not in self.checked]
        self.layoutChanged.emit()


class SideviewListModel(UniversalListModel):
    """List model for displaying location side view for hovered/selected location."""

    ObjectRole = Qt.UserRole + 1

    maxChanged = Signal()

    def __init__(
        self,
        object_class,
        objects=None,
        selected_objects=None,
        hovered_objects=None,
        parent=None,
    ):
        """Initialize list model.

        Args:
            object_class (cls): class which wraps objects stored in list (TabItem,...)
            objects (dict[str, object]): dictionary of objects stored in list
            selected_objects (list[str]): list of ids (keys) of selected objects,
                selected objects are displayed in the list
            hovered_objects (list[str]): list of ids (keys) of objects that are hovered
            parent (object): required param for initialization
        """
        super(SideviewListModel, self).__init__(parent)
        self._object_class = object_class
        self._selected = [] if selected_objects is None else selected_objects
        self._hovered = [] if hovered_objects is None else hovered_objects
        self._selected_max_heat = 1
        self._hovered_max_heat = 1
        self._is_hovered = False

        if objects is None:
            self._objects = {}
        else:
            self._objects = {k: self._object_class(v) for k, v in objects.items()}

    def set_selected(self, selected, check=False):
        """Set list of names of clicked locations."""
        self._selected = list(reversed(selected))
        if self._selected:
            self._selected_max_heat = max(self._objects[k].heat for k in self._selected)
            self.maxChanged.emit()
        self.layoutChanged.emit()

    def set_hovered(self, hovered, is_hovered=True):
        """Set list of names of currently hovered locations."""
        self._is_hovered = is_hovered
        self._hovered = list(reversed(hovered))
        if self._hovered:
            self._hovered_max_heat = max(self._objects[k].heat for k in self._hovered)
            self.maxChanged.emit()
        self.layoutChanged.emit()

    @Property(float, constant=False, notify=maxChanged)
    def max_heat(self):
        """Get current max heat of selected/hovered location."""
        return self._hovered_max_heat if self._is_hovered else self._selected_max_heat

    @Slot()
    def update(self):
        """Update maximal value, emit maxChanged signal which results in list redraw."""
        if self._selected:
            self._selected_max_heat = max(self._objects[k].heat for k in self._selected)
            self.maxChanged.emit()

    def rowCount(self, parent=QModelIndex()):
        """Get number of items in list (required method)."""
        if self._is_hovered:
            return len(self._hovered)
        return len(self._selected)

    def data(self, index, role):
        """Return data for given role and index (required method)."""
        if index.isValid() and role == UniversalListModel.ObjectRole:
            if self._is_hovered:
                return self._objects[self._hovered[index.row()]]
            return self._objects[self._selected[index.row()]]
        return None



"""Data model using Owlready2 ontology."""
import datetime
from subprocess import DEVNULL, check_call
from typing import List

import owlready2
from owlready2 import (
    ConstrainedDatatype,
    DataProperty,
    FunctionalProperty,
    ObjectProperty,
    Thing,
    default_world,
    destroy_entity,
    get_ontology,
    sync_reasoner,
    sync_reasoner_pellet,
)
from PySide2.QtCore import Property, QObject, Signal, Slot

from virtual_warehouse.data.utils import (
    convert_date,
    convert_dim,
    convert_type,
    convert_weight,
)

# We could use actual irl with already created ontology
BASE_IRI = "http://warehouse/onto.owl"
onto = get_ontology(BASE_IRI)


def destroy_all(cls):
    """Destroy all instances of given class.

    Args:
        cls (class): ontology class
    """
    for i in cls.instances():
        destroy_entity(i)


def save_ontology(file_path):
    """Save ontology in RDF/XML format.

    Args:
        file_path (str): file url should be processed with QUrl(file_path)
    """
    onto.save(file_path, format="rdfxml")


class OntoManager(QObject):
    """Class controlling ontology classes and reasoning."""

    classesChanged = Signal()
    javaChanged = Signal()

    def __init__(self):
        """Initialize OntoController."""
        QObject.__init__(self)
        self._classes = {}
        self._check_java()

    @Property(str, constant=False, notify=javaChanged)
    def java(self):
        """Get Java executable path for owlready reasoner."""
        return owlready2.JAVA_EXE

    @java.setter
    def set_java(self, val):
        """Set Java executable path for owlready reasoner."""
        owlready2.JAVA_EXE = val
        self._check_java()
        self.javaChanged.emit()

    @Property(bool, constant=False, notify=javaChanged)
    def java_correct(self):
        """Get true java path is correct."""
        return self._java_correct

    def _check_java(self):
        """Check if java path is set correctly."""
        try:
            check_call(
                [owlready2.JAVA_EXE, "--version"], stdout=DEVNULL, stderr=DEVNULL
            )
            self._java_correct = True
        except Exception:
            self._java_correct = False

    @Slot(str)
    def delete_class(self, cls):
        """Destroy given ontology class.

        Args:
            cls (str): name of new class for destruction
        """
        if cls in self._classes:
            destroy_entity(self._classes.pop(cls)[0])
            self.classesChanged.emit()

    def get_instances(self, cls):
        """Get instances of custom class.

        Args:
            cls (str): name of new classes

        Returns:
            list[Object]: list of class (RackLocation/Item/Order) instances
        """
        return self._classes[cls][0].instances()

    @Property("QVariantList", constant=False, notify=classesChanged)
    def classes(self):
        """Get list of classes for displaying in sideview."""
        return [
            {"name": k, "class": v[1], "count": len(v[0].instances())}
            for k, v in self._classes.items()
        ]

    @Slot(str, str, str)
    def create_class(self, name, cls, conditions):
        """Construct new ontology class based on RackLocation, Item or Order.

        Args:
            name (str): name of new class
            cls (str): string describing class, possible values: "RackLocation", "Item", "Order"
            conditions (str): describing condition (later replace by more complex structure)
        """
        # TODO: Checks, errors, background thread + loading
        assert cls in ["RackLocation", "Item", "Order"], "Invalid class type"

        if len(conditions.strip()) != 0:
            full_condition = f"[{cls} & {conditions}]"
        else:
            full_condition = f"[{cls}]"

        with onto:
            new_class = type(
                name, (eval(cls),), {"equivalent_to": eval(full_condition)}
            )
            sync_reasoner_pellet(debug=0, infer_property_values=False)
        self._classes[name] = (new_class, cls)
        self.classesChanged.emit()
        return new_class


with onto:

    class Country(Thing):
        """Class representing different countries.

        Attributes:
            name (str): Name of the country
        """

        pass

    class PhysicalObject(Thing):
        """Class representing physical object with dimensions.

        Attributes:
            has_length (float): length of packaging (in meters)
            has_width (float): width of packaging (in meters)
            has_height (float): height of packaging (in meters)
        """

    class ItemInstance(Thing):
        """Class instantiating item in some quantity.

        Attributes:
            has_item (Item): instantiated item
        """

    class Inventory(ItemInstance):
        """Description of inventory balance instance for given (date, location, item).

        Attributes:
            name (str): unique ID of inventory object
            has_date (datetime.datetime): date of the inventory status
            has_location (RackLocation): location object related to this inventory status
            # has_ltype (str): location type
            has_item (Item): item on given location
            has_expiry_date (datetime.datetime): date of the item expiration
            has_available_qty (int): available quantity of the item
            has_onhand_qty (int): on-hand quantity of the item
            has_transit_qty (int): transiting quantity of the item
            has_allocated_qty (int): allocated quantity of the item
            has_suspense_qty (int): suspended quantity of the item
        """

        @classmethod
        def create(
            cls,
            date: str,
            location_id: str,
            ltype: str,
            item_id: str,
            expiry_date: str,
            available_qty: int,
            onhand_qty: int,
            transit_qty: int,
            allocated_qty: int,
            suspense_qty: int,
        ):
            date_t = convert_date(date, "%d.%m.%Y")
            expiry_date = convert_date(expiry_date, "%d.%m.%Y")

            item = onto.search_one(iri=f"{BASE_IRI}#{item_id}")
            location = onto.search_one(iri=f"{BASE_IRI}#{location_id}")

            return cls(
                f"{date.strftime('%Y:%m:%d')}-{location_id}-{item_id}",
                has_date=date_t,
                has_location=location,
                # has_ltype=ltype,
                has_item=item,
                has_expiry_date=expiry_date,
                has_available_qty=available_qty,
                has_onhand_qty=onhand_qty,
                has_transit_qty=transit_qty,
                has_allocated_qty=allocated_qty,
                suspense_qty=suspense_qty,
            )

        @classmethod
        def destroy_all(cls):
            """Destroy all instances of Inventory class."""
            destroy_all(cls)

        @staticmethod
        def get_by_item(item, date):
            """Get list of inventories containing item on given date.

            Args:
                item (Item): item object to look for.
                date (datetime.datetime): date of inventory

            Returns:
                List[Item]: list of items stored at locations
            """
            return onto.search(type=Inventory, has_item=item)

    class ItemUnit(PhysicalObject):
        """Represents different packaging units of item.

        Attributes:
            name (str): unique id (item_id-unit_lvl)
            has_conversion_qty (int): number of packages inside base unit
            has_ref_qty_uom (str): name of reference quantity unit of measure
            has_length (float): length of packaging (in meters)
            has_width (float): width of packaging (in meters)
            has_height (float): height of packaging (in meters)
            has_weight (float): weight of packaging (in kg)
        """

        @classmethod
        def create(
            cls,
            _id: str,
            conversion_qty: int,
            qty_uom: str,
            length: float,
            width: float,
            height: float,
            dim_uom: str,
            weight: float,
            weight_uom: str,
        ):
            length = convert_dim(length, dim_uom)
            width = convert_dim(width, dim_uom)
            height = convert_dim(height, dim_uom)
            weight = convert_weight(weight, weight_uom)

            return cls(
                _id,
                has_conversion_qty=conversion_qty,
                has_ref_qty_uom=qty_uom,
                has_length=length,
                has_width=width,
                has_height=height,
                has_weight=weight,
            )

    class Item(Thing):
        """Basic description of item, including packaging units.

        Attributes:
            name (str): id of item
            has_description (str): description of items
            has_gtype (str): type of goods
            has_required_zone (str): required zone for storing item
            has_base_unit (ItemUnit): base packaging unit
            has_unit_levels (List[ItemUnit]): list of all different packaging units
        """

        @classmethod
        def create(
            cls,
            _id: str,
            description: str,
            gtype: str,
            zone: str,
            base_unit: ItemUnit,
            unit_levels: List[ItemUnit],
        ):
            return cls(
                _id,
                has_description=description,
                has_gtype=gtype,
                has_required_zone=zone,
                has_base_unit=base_unit,
                has_unit_levels=unit_levels,
            )

        @classmethod
        def destroy_all(cls):
            """Destroy all instances of Item class as well as related entities."""
            destroy_all(cls)
            destroy_all(ItemUnit)

        @staticmethod
        def get_by_locations(locations):
            """Get list of items stored at given location.

            Args:
                locations (List[RackLocation]): list of locations to look for.

            Returns:
                List[Item]: list of items stored at locations
            """
            # TODO: Check inventory date
            values = " ".join(map(lambda x: f"<{x.iri}>", locations))
            i1 = "VALUES ?loc { " + values + " } .\n"
            i2 = f"?inv <{has_location.iri}> ?loc .\n"
            i3 = f"?inv <{has_item.iri}> ?item .\n"

            query = "SELECT DISTINCT ?item WHERE { " + i1 + i2 + i3 + " }"
            return list(default_world.sparql_query(query))

        @staticmethod
        def get_by_orders(orders):
            """Get list of items included in given orders.

            Args:
                orders (List[Order]): list of orders to search.

            Returns:
                List[Item]: list of items included in orders.
            """
            values = " ".join(map(lambda x: f"<{x.iri}>", orders))
            i1 = "VALUES ?ord { " + values + " } .\n"
            i2 = f"?ord <{has_ordered_items.iri}> ?oi .\n"
            i3 = f"?oi <{has_item.iri}> ?item .\n"

            query = "SELECT DISTINCT ?item WHERE { " + i1 + i2 + i3 + " }"
            return list(default_world.sparql_query(query))

    class Location(PhysicalObject):
        """Description of location in warehouse.

        Attributes:
            name (str): unique id of location
            has_ltype (str): type of location
            has_lclass (str): class of location
            has_lsubclass (str): subclass of location
            has_length (float): length of location
            has_width (float): width of location
            has_height (float): height of location
            has_max_weight (float): maximal weight which can location hold
            has_zone (str): zone where is location located
            has_x (float): x coordinate of location
            has_y (float): y coordinate of location
            has_z (float): z coordinate of location
        """

        @classmethod
        def create(
            cls,
            _id: str,
            ltype: str,
            lclass: str,
            lsubclass: str,
            length: float,
            width: float,
            height: float,
            dim_uom: str = None,
            max_weight: float = None,
            weight_uom: str = None,
            zone: str = None,
            # Params which are usually calculated later:
            x: float = None,
            y: float = None,
            z: float = None,
        ):
            ltype = convert_type(ltype)
            length = convert_dim(length, dim_uom)
            width = convert_dim(width, dim_uom)
            height = convert_dim(height, dim_uom)
            if max_weight:
                max_weight = convert_weight(max_weight, weight_uom)

            return cls(
                _id,
                has_ltype=ltype,
                has_lclass=lclass,
                has_lsubclass=lsubclass,
                has_length=length,
                has_width=width,
                has_height=height,
                has_max_weight=max_weight,
                has_zone=zone,
                has_x=x,
                has_y=y,
                has_z=z,
            )

        def set_coord(self, x: float, y: float, z: float):
            """Additionally set coordinates of the location."""
            self.has_x, self.has_y, self.has_z = x, y, z  # skipcq: PYL-W0201

        def get_2d(self):
            """Get planar coordinates of location (used for top-down view)."""
            return (self.has_x, self.has_y)

        @classmethod
        def destroy_all(cls):
            """Destroy all instances of Location class."""
            destroy_all(cls)

    class RackLocation(Location):
        """Description of location which can store items.

        Attributes:
            name (str): unique id of location
            has_ltype (str): type of location
            has_lclass (str): class of location
            has_lsubclass (str): subclass of location
            has_length (float): length of location
            has_width (float): width of location
            has_height (float): height of location
            has_max_weight (float): maximal weight which can location hold
            has_zone (str): zone where is location located
            has_x (float): x coordinate of location
            has_y (float): y coordinate of location
            has_z (float): z coordinate of location
            has_freq (float): frequency calculated for heat map displaying
        """

        has_freq: float = 0

        @staticmethod
        def get_by_orders(orders):
            """Get list of locations which are potentially accessed by given orders.

            Args:
                locations (List[Order]): list of locations to inspect

            Returns:
                List[RackLocation]: list of locations containing ordered items
            """
            # TODO: speed up this query (reverse property???)
            values = " ".join(map(lambda x: f"<{x.iri}>", orders))
            i1 = "VALUES ?ord { " + values + " } .\n"
            i2 = f"?ord <{has_ordered_items.iri}> ?oi .\n"
            i3 = f"?oi <{has_item.iri}> ?item .\n"
            i4 = f"?inv <{has_item.iri}> ?item .\n"
            i5 = f"?inv <{has_location.iri}> ?loc .\n"

            query = "SELECT DISTINCT ?loc WHERE { " + i1 + i2 + i3 + i4 + i5 + " }"
            return list(default_world.sparql_query(query))

        @staticmethod
        def get_by_items(items):
            """Get list of locations storing given items.

            Args:
                items (List[Item]): list of items to locate items.

            Returns:
                List[RackLocation]: list of locations storing items
            """
            values = " ".join(map(lambda x: f"<{x.iri}>", items))
            i1 = "VALUES ?item { " + values + " } .\n"
            i2 = f"?inv <{has_item.iri}> ?item .\n"
            i3 = f"?inv <{has_location.iri}> ?loc .\n"

            query = "SELECT DISTINCT ?loc WHERE { " + i1 + i2 + i3 + " }"
            return list(default_world.sparql_query(query))

    class OrderedItem(ItemInstance):
        """Description of item instance in order.

        Attributes:
            name (str): id of object (ordered_id-item_id)
            has_item (Item): item object instantiated by this order
            has_requested_qty (int): requested amount of item inside ordered
            has_total_qty (int): amount of item provided inside order
            has_qty_uom (str): unit of measure of quantities
        """

        pass

    class Order(Thing):
        """Description of single order from warehouse.

        Attributes:
            name (str): id of the order
            has_direction (str): outbound/inbound
            has_country (str): Country object representing shipping destination
            has_delivery_date (datetime.datetime): requested delivery date
            has_s_ship_date (datetime.datetime): scheduled sipping date
            has_a_ship_date (datetime.datetime): actual sipping date
            has_line_num (int): ERP order line number
            has_ordered_items (List[OrderedItem]): list of item instances inside order
        """

        @classmethod
        def create(
            cls,
            _id: str,
            direction: str,
            country_id: str,
            delivery_date: str,
            s_ship_date: str,
            a_ship_date: str,
            line_num: int,
            item_id: str,
            requested_qty: int,
            total_qty: int,
            qty_uom: str,
        ):
            _id = str(_id)
            # Convert to datetime
            delivery_date = convert_date(delivery_date, "%d.%m.%Y")
            s_ship_date = convert_date(s_ship_date, "%d.%m.%Y")
            a_ship_date = convert_date(a_ship_date, "%d.%m.%Y")

            country = onto.search_one(iri=f"{BASE_IRI}#{country_id}")
            if country is None:
                country = Country(country_id)

            item = onto.search_one(iri=f"{BASE_IRI}#{item_id}")
            oi = OrderedItem(
                f"{_id}-{item_id}",
                has_item=item,
                has_requested_qty=requested_qty,
                has_total_qty=total_qty,
                has_qty_uom=qty_uom,
            )

            return cls(
                _id,
                has_direction=direction,
                has_country=country,
                has_delivery_date=delivery_date,
                has_s_ship_date=s_ship_date,
                has_a_ship_date=a_ship_date,
                has_line_num=line_num,
                has_ordered_items=[oi],
            )

        def add_item(
            self, item_id: str, requested_qty: int, total_qty: int, qty_uom: str
        ):
            """Create and add item instance into order."""
            item = onto.search_one(iri=f"{BASE_IRI}#{item_id}")
            oi = OrderedItem(
                f"{self.name}-{item_id}",
                has_item=item,
                has_requested_qty=requested_qty,
                has_total_qty=total_qty,
                has_qty_uom=qty_uom,
            )
            self.has_ordered_items.append(oi)

        @classmethod
        def destroy_all(cls):
            """Destroy all instances of Order class as well as related entities."""
            destroy_all(cls)
            destroy_all(OrderedItem)

        @staticmethod
        def get_by_items(items):
            """Get list of orders which contains given items.

            Args:
                items (List[Item]): list of items to look for.

            Returns:
                List[Order]: list of orders containing at leas one of the provided items
            """
            values = " ".join(map(lambda x: f"<{x.iri}>", items))
            i1 = "VALUES ?item { " + values + " } .\n"
            i2 = f"?oi <{has_item.iri}> ?item .\n"
            i3 = f"?ord <{has_ordered_items.iri}> ?oi .\n"

            query = "SELECT DISTINCT ?ord WHERE { " + i1 + i2 + i3 + " }"
            return list(default_world.sparql_query(query))

        @staticmethod
        def get_by_locations(locations):
            """Get list of orders which potentially access given locations.

            Args:
                locations (List[RackLocation]): list of locations to inspect

            Returns:
                List[Order]: list of orders containing items stored at given locations
            """
            values = " ".join(map(lambda x: f"<{x.iri}>", locations))
            i1 = "VALUES ?loc { " + values + " } .\n"
            i2 = f"?inv <{has_location.iri}> ?loc .\n"
            i3 = f"?inv <{has_item.iri}> ?item .\n"
            i4 = f"?oi <{has_item.iri}> ?item .\n"
            i5 = f"?ord <{has_ordered_items.iri}> ?oi .\n"

            query = "SELECT DISTINCT ?ord WHERE { " + i1 + i2 + i3 + i4 + i5 + " }"
            return list(default_world.sparql_query(query))

    # Order properties
    class has_direction(Order >> str, FunctionalProperty):
        """Direction of Order."""

    class has_country(Order >> Country, FunctionalProperty):
        """Country of destination of Order."""

    class has_delivery_date(Order >> datetime.datetime, FunctionalProperty):
        """Delivery date of Order."""

    class has_s_ship_date(Order >> datetime.datetime, FunctionalProperty):
        """Scheduled shipping date of Order."""

    class has_a_ship_date(Order >> datetime.datetime, FunctionalProperty):
        """Actual shipping date of Order."""

    class has_line_num(Order >> int, FunctionalProperty):
        """Line number of Order."""

    class has_ordered_items(Order >> OrderedItem):
        """List of OrderedItem (instances of Item) inside Order."""

    # OrderedItem properties
    class has_item(ItemInstance >> Item, FunctionalProperty):
        """Item which is instantiated."""

    class in_order(OrderedItem >> Order, FunctionalProperty):
        """Pointer to Order which contains this OrderedItem."""

        inverse_property = has_ordered_items

    class has_requested_qty(OrderedItem >> int, FunctionalProperty):
        """Requested quantity of Item."""

    class has_total_qty(OrderedItem >> int, FunctionalProperty):
        """Total quantity of Item."""

    class has_qty_uom(OrderedItem >> str, FunctionalProperty):
        """Quantity unit of measure used for total_qty and requested_qty."""

    # ItemUnit properties
    class has_conversion_qty(ItemUnit >> int, FunctionalProperty):
        """Conversion quantity of ItemUnit."""

    class has_weight(ItemUnit >> float, FunctionalProperty):
        """Conversion quantity of ItemUnit."""

    class has_ref_qty_uom(ItemUnit >> str, FunctionalProperty):
        """Quantity unit of measure used for total_qty and requested_qty."""

    # Item properties
    class has_description(Item >> str, FunctionalProperty):
        """Description of Item."""

    class has_gtype(Item >> str, FunctionalProperty):
        """Good type of Item."""

    class has_required_zone(Item >> str, FunctionalProperty):
        """Zone required by Item."""

    class has_base_unit(Item >> ItemUnit, FunctionalProperty):
        """Base ItemUnit of Item."""

    class has_unit_levels(Item >> ItemUnit):
        """List of ItemUnits (unit levels) of Item."""

    # PhysicalObject properties
    class has_length(PhysicalObject >> float, FunctionalProperty):
        """Length of Object."""

    class has_width(PhysicalObject >> float, FunctionalProperty):
        """Width of Object."""

    class has_height(PhysicalObject >> float, FunctionalProperty):
        """Height of Object."""

    # Location properties
    class has_ltype(Location >> str, FunctionalProperty):
        """Location type."""

    class has_lclass(Location >> str, FunctionalProperty):
        """Location class."""

    class has_lsubclass(Location >> str, FunctionalProperty):
        """Location subclass."""

    class has_max_weight(Location >> float, FunctionalProperty):
        """Max weight which Location can hold."""

    # class has_freq(Location >> int, FunctionalProperty):
    #     """Frequency calculated for the location."""

    class has_x(Location >> float, FunctionalProperty):
        """x coordinate of Location."""

    class has_y(Location >> float, FunctionalProperty):
        """y coordinate of Location."""

    class has_z(Location >> float, FunctionalProperty):
        """z coordinate of Location."""

    class has_zone(Location >> str, FunctionalProperty):
        """Zone where is Location located."""

    # Inventory properties
    class has_date(Inventory >> datetime.datetime, FunctionalProperty):
        """Date of Inventory report."""

    class has_location(Inventory >> RackLocation, FunctionalProperty):
        """Location described by Inventory."""

    class has_expiry_date(Inventory >> datetime.datetime, FunctionalProperty):
        """Expiry date of Item in Inventory."""

    class has_available_qty(Inventory >> int, FunctionalProperty):
        """Available quantity of Item."""

    class has_onhand_qty(Inventory >> int, FunctionalProperty):
        """On-hand quantity of Item."""

    class has_transit_qty(Inventory >> int, FunctionalProperty):
        """Quantity in transit of Item."""

    class has_allocated_qty(Inventory >> int, FunctionalProperty):
        """Allocated quantity of Item."""

    class has_supsense_qty(Inventory >> int, FunctionalProperty):
        """Suspense quantity of Item."""


# sync_reasoner(infer_property_values=True)

# SQLite3
# default_world.graph.db
# default_world.graph.execute



"""Module with plug-in for calculating selected items frequencies."""

from virtual_warehouse.data.data_model import Inventory
from virtual_warehouse.plugin import BasePlugin


class Plugin(BasePlugin):
    """Sub-class of base plugin providing calculation of frequency.
    Calculate frequency as on-hand quantity of selected items.
    """

    display_name = "&Item Histogram"

    def __init__(self, locations, items, orders, inventory):
        """Initialize plugin and save required parameters.

        Args:
            locations (dict): dictionary of locations mapping id to locations
            items (dict): dictionary of items mapping id to items
            orders (dict): dictionary of orders mapping id to order
            inventory (dict): dictionary of specifying inventory
        """
        super().__init__(locations, items, orders, inventory)
        self.inventory = inventory
        self.items = items
        self.date = list(self.inventory.keys())[-1]

    def on_items_update(self, clear, add, ids):
        """Update frequency calculation on items check/uncheck.

        Args:
            clear (bool): Whether to clear previously selected orders.
            add (bool): Whether to add or remove ids from calculation.
            ids (list[str]): list of ids considered in calculation.
        """
        if clear:
            self._clear_frequencies()

        for item_id in ids:
            invs = Inventory.get_by_item(self.items[item_id], self.date)
            for inv in invs:
                inv.has_location.has_freq += (1 if add else -1) * inv.has_onhand_qty

    def _calculate_freq(self, locations, items, orders):
        """Calculate frequency for individual locations based on selected orders."""
        self.on_items_update(False, True, items)
